#include <ReWindow/types/Window.h>
#include <ReWindow/Logger.h>
#include <dlfcn.h>
#include <set>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <vulkan/vulkan.h>
#include <vulkan/vulkan_xlib.h>

using namespace ReWindow;
class ReWindow::RWindow::Platform {
public:
    Window window;
    XEvent xev;
    Display* display;
    int defaultScreen;
    XVisualInfo* visual;
    XSetWindowAttributes xSetWindowAttributes;
    XWindowAttributes windowAttributes;
    Atom wmDeleteWindow;
    Atom windowTypeAtom;
    Atom windowTypeUtilityAtom;
    XSizeHints hints;
    Cursor invisible_cursor = 0;
    XWMHints* wm_hints = nullptr;
};

namespace Vulkan {
    bool constructor_success = false;
    PFN_vkGetInstanceProcAddr vkGetInstanceProcAddr = nullptr;
    PFN_vkCreateInstance vkCreateInstance = nullptr;
    PFN_vkGetPhysicalDeviceProperties vkGetPhysicalDeviceProperties = nullptr;
    PFN_vkEnumeratePhysicalDevices vkEnumeratePhysicalDevices = nullptr;
    PFN_vkGetPhysicalDeviceQueueFamilyProperties vkGetPhysicalDeviceQueueFamilyProperties = nullptr;
    PFN_vkGetDeviceQueue vkGetDeviceQueue = nullptr;
    PFN_vkCreateDevice vkCreateDevice = nullptr;
    PFN_vkCreateXlibSurfaceKHR vkCreateXlibSurfaceKHR = nullptr;
    PFN_vkGetPhysicalDeviceSurfaceSupportKHR vkGetPhysicalDeviceSurfaceSupportKHR = nullptr;
}

void* vulkan_lib = nullptr;
VkInstance vulkan_context = nullptr;
VkApplicationInfo vulkan_application_info{};
VkSurfaceKHR vulkan_surface = nullptr;
VkDevice vulkan_logical_device = nullptr;
VkQueue vulkan_graphics_queue = nullptr;
VkQueue vulkan_present_queue = nullptr;

// Necessary because some functions aren't available until you have a context (Vulkan is weird).
void LoadVulkanInstancedFunctions(VkInstance vulkan_ctx) {
    Vulkan::vkGetPhysicalDeviceProperties = (PFN_vkGetPhysicalDeviceProperties) Vulkan::vkGetInstanceProcAddr(vulkan_ctx, "vkGetPhysicalDeviceProperties");
    Vulkan::vkEnumeratePhysicalDevices = (PFN_vkEnumeratePhysicalDevices) Vulkan::vkGetInstanceProcAddr(vulkan_ctx, "vkEnumeratePhysicalDevices");
    Vulkan::vkGetPhysicalDeviceQueueFamilyProperties = (PFN_vkGetPhysicalDeviceQueueFamilyProperties) Vulkan::vkGetInstanceProcAddr(vulkan_ctx, "vkGetPhysicalDeviceQueueFamilyProperties");
    Vulkan::vkGetDeviceQueue = (PFN_vkGetDeviceQueue) Vulkan::vkGetInstanceProcAddr(vulkan_ctx, "vkGetDeviceQueue");
    Vulkan::vkCreateDevice = (PFN_vkCreateDevice) Vulkan::vkGetInstanceProcAddr(vulkan_ctx, "vkCreateDevice");
    Vulkan::vkCreateXlibSurfaceKHR = (PFN_vkCreateXlibSurfaceKHR) Vulkan::vkGetInstanceProcAddr(vulkan_ctx, "vkCreateXlibSurfaceKHR");
    Vulkan::vkGetPhysicalDeviceSurfaceSupportKHR = (PFN_vkGetPhysicalDeviceSurfaceSupportKHR) Vulkan::vkGetInstanceProcAddr(vulkan_ctx, "vkGetPhysicalDeviceSurfaceSupportKHR");
}

bool VulkanWindow::Open() {
    if (!Vulkan::constructor_success)
        return false;

    platform->display = XOpenDisplay(nullptr);
    platform->defaultScreen = DefaultScreen(platform->display);

    platform->xSetWindowAttributes.border_pixel = BlackPixel(platform->display, platform->defaultScreen);
    platform->xSetWindowAttributes.background_pixel = BlackPixel(platform->display, platform->defaultScreen);
    platform->xSetWindowAttributes.override_redirect = True;
    platform->xSetWindowAttributes.event_mask = ExposureMask;

    vulkan_application_info.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO;
    vulkan_application_info.pApplicationName = title.c_str();
    vulkan_application_info.applicationVersion = VK_MAKE_VERSION(1, 0, 0);
    vulkan_application_info.pEngineName = "";
    vulkan_application_info.engineVersion = VK_MAKE_VERSION(1, 0, 0);

    vulkan_application_info.apiVersion = VK_API_VERSION_1_0;

    VkInstanceCreateInfo create_info{};
    create_info.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
    create_info.pApplicationInfo = &vulkan_application_info;

    const char* extensions[] = { "VK_KHR_surface", "VK_KHR_xlib_surface" };
    create_info.enabledExtensionCount = 2;
    create_info.ppEnabledExtensionNames = extensions;

    if (Vulkan::vkCreateInstance(&create_info, nullptr, &vulkan_context) != VK_SUCCESS) {
        Logger::Error("Couldn't create Vulkan instance.");
        return false;
    }

    LoadVulkanInstancedFunctions(vulkan_context);
    XVisualInfo visual_info_t{};
    visual_info_t.screen = platform->defaultScreen;
    int visual_count;
    XVisualInfo* vi = XGetVisualInfo(platform->display, VisualScreenMask, &visual_info_t, &visual_count);

    if (!vi)
        return false;

    Colormap colormap = XCreateColormap(platform->display, RootWindow(platform->display, vi->screen), vi->visual, AllocNone);
    platform->xSetWindowAttributes.colormap = colormap;

    platform->window = XCreateWindow(platform->display, RootWindow(platform->display, vi->screen),
                                 0, 0, width, height, 0, vi->depth, InputOutput,
                                 vi->visual, CWBackPixel | CWColormap | CWBorderPixel, &platform->xSetWindowAttributes);

    if (!platform->window) {
        Logger::Error("Failed to create X11 window for Vulkan.");
        return false;
    }

    XFree(vi);

    VkXlibSurfaceCreateInfoKHR surface_create_info{};
    surface_create_info.sType = VK_STRUCTURE_TYPE_XLIB_SURFACE_CREATE_INFO_KHR;
    surface_create_info.dpy = platform->display;
    surface_create_info.window = platform-> window;

    if (Vulkan::vkCreateXlibSurfaceKHR(vulkan_context, &surface_create_info, nullptr, &vulkan_surface) != VK_SUCCESS) {
        Logger::Error("Failed to create Vulkan X11 surface.");
        return false;
    }

    unsigned int device_count = 0;
    Vulkan::vkEnumeratePhysicalDevices(vulkan_context, &device_count, nullptr);

    if (device_count == 0) {
        Logger::Error("Opening the window as Vulkan but there isn't a Vulkan compatible graphics card?");
        return false;
    }

    std::vector<VkPhysicalDevice> vulkan_devices(device_count);
    Vulkan::vkEnumeratePhysicalDevices(vulkan_context, &device_count, vulkan_devices.data());

    VkPhysicalDevice selected_device = VK_NULL_HANDLE;

    int graphics_family = -1;
    int present_family = -1;
    for (const auto& d : vulkan_devices) {
        unsigned int queue_family_count = 0;
        Vulkan::vkGetPhysicalDeviceQueueFamilyProperties(d, &queue_family_count, nullptr);

        std::vector<VkQueueFamilyProperties> queue_families(queue_family_count);
        Vulkan::vkGetPhysicalDeviceQueueFamilyProperties(d, &queue_family_count, queue_families.data());

        bool has_graphics = false;
        bool has_present = false;
        for (unsigned int i = 0; i < queue_family_count; i++) {
            if (queue_families[i].queueFlags & VK_QUEUE_GRAPHICS_BIT) { has_graphics = true; graphics_family = i; }

            VkBool32 present_support = false;
            Vulkan::vkGetPhysicalDeviceSurfaceSupportKHR(d, i, vulkan_surface, &present_support);

            if (present_support) { has_present = true; present_family = i; }

            if (has_graphics && has_present) { selected_device = d; break; }
        }
        if (selected_device != VK_NULL_HANDLE)
            break;
    }

    if (selected_device == VK_NULL_HANDLE) {
        Logger::Error("There isn't a Vulkan device which supports rendering & presenting?");
        return false;
    }

    VkPhysicalDeviceProperties device_properties;
    Vulkan::vkGetPhysicalDeviceProperties(selected_device, &device_properties);
    Logger::Debug("Vulkan continuing with selected device: " + std::string(device_properties.deviceName));

    float queue_prio = 1.0f;
    std::vector<VkDeviceQueueCreateInfo> queue_create_info;
    std::set<unsigned int> unique_queue_families = { (unsigned int) graphics_family, (unsigned int) present_family};

    for (unsigned int qf : unique_queue_families) {
        VkDeviceQueueCreateInfo vdqci{};
        vdqci.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
        vdqci.queueFamilyIndex = qf;
        vdqci.queueCount = 1;
        vdqci.pQueuePriorities = &queue_prio;
        queue_create_info.push_back(vdqci);
    }

    // TODO allow the user to specify this.
    VkPhysicalDeviceFeatures enabled_features{};

    VkDeviceCreateInfo device_create_info{};
    device_create_info.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
    device_create_info.queueCreateInfoCount = queue_create_info.size();
    device_create_info.pQueueCreateInfos = queue_create_info.data();
    device_create_info.pEnabledFeatures = &enabled_features;

    // TODO allow the user to specify this.
    std::vector<const char*> device_ext = {VK_KHR_SWAPCHAIN_EXTENSION_NAME};
    device_create_info.enabledExtensionCount = device_ext.size();
    device_create_info.ppEnabledExtensionNames = device_ext.data();

    if (Vulkan::vkCreateDevice(selected_device, &device_create_info, nullptr, &vulkan_logical_device) != VK_SUCCESS) {
        Logger::Error("Couldn't create the Vulkan logical device.");
        return false;
    }

    Vulkan::vkGetDeviceQueue(vulkan_logical_device, graphics_family, 0, &vulkan_graphics_queue);
    Vulkan::vkGetDeviceQueue(vulkan_logical_device, present_family, 0, &vulkan_present_queue);





    XSelectInput(platform->display, platform->window, ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
                                              PointerMotionHintMask | FocusChangeMask | StructureNotifyMask | SubstructureRedirectMask | SubstructureNotifyMask | CWColormap );
    XMapWindow(platform->display, platform->window);
    XStoreName(platform->display, platform->window, title.c_str());

    platform->windowTypeAtom = XInternAtom(platform->display, "_NET_WM_WINDOW_TYPE", False);
    platform->windowTypeUtilityAtom = XInternAtom(platform->display, "_NET_WM_WINDOW_TYPE_UTILITY", False);
    XChangeProperty(platform->display, platform->window, platform->windowTypeAtom, XA_ATOM, 32, PropModeReplace, (unsigned char*) &platform->windowTypeUtilityAtom, 1);

    platform->wmDeleteWindow = XInternAtom(platform->display, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(platform->display, platform->window, &platform->wmDeleteWindow, 1);

    XGetWindowAttributes(platform->display, platform->window, &platform->windowAttributes);
    render_area_position = { platform->windowAttributes.x, platform->windowAttributes.y };

    open = true;
    processOnOpen();
    return true;
}

void VulkanWindow::SwapBuffers() {

}

void VulkanWindow::SetVsyncEnabled(bool state) {

}

std::string VulkanWindow::GetGraphicsDriverVendor() {

}

VulkanWindow::VulkanWindow(const std::string& title, int width, int height) : RWindow(title, width, height) {
    vulkan_lib = dlopen("libvulkan.so", RTLD_NOW | RTLD_LOCAL);
    if (!vulkan_lib)
        vulkan_lib = dlopen("libvulkan.so.1", RTLD_NOW | RTLD_LOCAL);
    if (!vulkan_lib)
        Logger::Debug("libvulkan.so couldn't be found in your LD_LIBRARY_PATH."),
        Vulkan::constructor_success = false;

    Vulkan::vkGetInstanceProcAddr = (PFN_vkGetInstanceProcAddr) dlsym(vulkan_lib,"vkGetInstanceProcAddr");
    Vulkan::vkCreateInstance = (PFN_vkCreateInstance) Vulkan::vkGetInstanceProcAddr(NULL, "vkCreateInstance");
    Vulkan::constructor_success = true;
}